package javax.swing;

import java.awt.event.ActionEvent;
import java.util.Iterator;
import java.awt.event.ActionListener;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.lang.ref.WeakReference;
import java.beans.PropertyChangeEvent;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.LinkedBlockingQueue;
import sun.awt.AppContext;
import java.util.concurrent.ExecutorService;
import java.beans.PropertyChangeListener;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.ExecutionException;
import java.util.List;
import java.util.concurrent.Callable;
import sun.swing.AccumulativeRunnable;
import java.beans.PropertyChangeSupport;
import java.util.concurrent.FutureTask;
import java.util.concurrent.RunnableFuture;

public abstract class SwingWorker<T, V> implements RunnableFuture<T>
{
    private static final int MAX_WORKER_THREADS = 10;
    private volatile int progress;
    private volatile SwingWorker.StateValue state;
    private final FutureTask<T> future;
    private final PropertyChangeSupport propertyChangeSupport;
    private AccumulativeRunnable<V> doProcess;
    private AccumulativeRunnable<Integer> doNotifyProgressChange;
    private final AccumulativeRunnable<Runnable> doSubmit;
    private static final Object DO_SUBMIT_KEY;
    
    public SwingWorker() {
        this.doSubmit = getDoSubmit();
        this.future = new FutureTask<T>(new Callable<T>() {
            @Override
            public T call() throws Exception {
                SwingWorker.access$000(SwingWorker.this, SwingWorker.StateValue.STARTED);
                return SwingWorker.this.doInBackground();
            }
        }) {
            @Override
            protected void done() {
                SwingWorker.access$100(SwingWorker.this);
                SwingWorker.access$000(SwingWorker.this, SwingWorker.StateValue.DONE);
            }
        };
        this.state = SwingWorker.StateValue.PENDING;
        this.propertyChangeSupport = new SwingWorker.SwingWorkerPropertyChangeSupport(this);
        this.doProcess = null;
        this.doNotifyProgressChange = null;
    }
    
    protected abstract T doInBackground() throws Exception;
    
    @Override
    public final void run() {
        this.future.run();
    }
    
    @SafeVarargs
    protected final void publish(V... args) {
        synchronized (this) {
            if (this.doProcess == null) {
                this.doProcess = new AccumulativeRunnable<V>() {
                    public void run(List<V> list) {
                        SwingWorker.this.process(list);
                    }
                    
                    @Override
                    protected void submit() {
                        SwingWorker.this.doSubmit.add(new Runnable[] { this });
                    }
                };
            }
        }
        this.doProcess.add(args);
    }
    
    protected void process(List<V> list) {
    }
    
    protected void done() {
    }
    
    protected final void setProgress(int n) {
        if (n < 0 || n > 100) {
            throw new IllegalArgumentException("the value should be from 0 to 100");
        }
        if (this.progress == n) {
            return;
        }
        int progress = this.progress;
        this.progress = n;
        if (!this.getPropertyChangeSupport().hasListeners("progress")) {
            return;
        }
        synchronized (this) {
            if (this.doNotifyProgressChange == null) {
                this.doNotifyProgressChange = new AccumulativeRunnable<Integer>() {
                    public void run(List<Integer> list) {
                        SwingWorker.this.firePropertyChange("progress", list.get(0), list.get(list.size() - 1));
                    }
                    
                    @Override
                    protected void submit() {
                        SwingWorker.this.doSubmit.add(new Runnable[] { this });
                    }
                };
            }
        }
        this.doNotifyProgressChange.add(new Integer[] { Integer.valueOf(progress), Integer.valueOf(n) });
        return;
    }
    
    public final int getProgress() {
        return this.progress;
    }
    
    public final void execute() {
        getWorkersExecutorService().execute(this);
    }
    
    @Override
    public final boolean cancel(boolean mayInterruptIfRunning) {
        return this.future.cancel(mayInterruptIfRunning);
    }
    
    @Override
    public final boolean isCancelled() {
        return this.future.isCancelled();
    }
    
    @Override
    public final boolean isDone() {
        return this.future.isDone();
    }
    
    @Override
    public final T get() throws InterruptedException, ExecutionException {
        return this.future.get();
    }
    
    @Override
    public final T get(long timeout, TimeUnit unit) throws InterruptedException, ExecutionException, TimeoutException {
        return this.future.get(timeout, unit);
    }
    
    public final void addPropertyChangeListener(PropertyChangeListener listener) {
        this.getPropertyChangeSupport().addPropertyChangeListener(listener);
    }
    
    public final void removePropertyChangeListener(PropertyChangeListener listener) {
        this.getPropertyChangeSupport().removePropertyChangeListener(listener);
    }
    
    public final void firePropertyChange(String propertyName, Object oldValue, Object newValue) {
        this.getPropertyChangeSupport().firePropertyChange(propertyName, oldValue, newValue);
    }
    
    public final PropertyChangeSupport getPropertyChangeSupport() {
        return this.propertyChangeSupport;
    }
    
    public final SwingWorker.StateValue getState() {
        if (this.isDone()) {
            return SwingWorker.StateValue.DONE;
        }
        return this.state;
    }
    
    private void setState(SwingWorker.StateValue state) {
        this.firePropertyChange("state", this.state, this.state = state);
    }
    
    private void doneEDT() {
        Runnable runnable = new Runnable() {
            @Override
            public void run() {
                SwingWorker.this.done();
            }
        };
        if (SwingUtilities.isEventDispatchThread()) {
            runnable.run();
        }
        else {
            this.doSubmit.add(new Runnable[] { runnable });
        }
    }
    
    private static synchronized ExecutorService getWorkersExecutorService() {
        AppContext appContext = AppContext.getAppContext();
        final ExecutorService value = (ExecutorService)appContext.get(SwingWorker.class);
        if (value == null) {
            value = new ThreadPoolExecutor(10, 10, 10L, TimeUnit.MINUTES, new LinkedBlockingQueue(), new ThreadFactory() {
                final ThreadFactory defaultFactory = Executors.defaultThreadFactory();
                
                @Override
                public Thread newThread(Runnable runnable) {
                    Thread thread = this.defaultFactory.newThread(runnable);
                    thread.setName("SwingWorker-" + thread.getName());
                    thread.setDaemon(true);
                    return thread;
                }
            });
            appContext.put(SwingWorker.class, value);
            appContext.addPropertyChangeListener("disposed", new PropertyChangeListener() {
                @Override
                public void propertyChange(PropertyChangeEvent propertyChangeEvent) {
                    if ((boolean)propertyChangeEvent.getNewValue()) {
                        ExecutorService executorService = (ExecutorService)new WeakReference(value).get();
                        if (executorService != null) {
                            AccessController.doPrivileged(new PrivilegedAction<Void>() {
                                @Override
                                public Void run() {
                                    executorService.shutdown();
                                    return null;
                                }
                            });
                        }
                    }
                }
            });
            return value;
        }
        return value;
    }
    
    private static AccumulativeRunnable<Runnable> getDoSubmit() {
        synchronized (SwingWorker.DO_SUBMIT_KEY) {
            AppContext appContext = AppContext.getAppContext();
            Object value = appContext.get(SwingWorker.DO_SUBMIT_KEY);
            if (value == null) {
                value = new SwingWorker.DoSubmitAccumulativeRunnable();
                appContext.put(SwingWorker.DO_SUBMIT_KEY, value);
            }
            return (AccumulativeRunnable)value;
        }
    }
    
    static void access$000(SwingWorker swingWorker, SwingWorker.StateValue state) {
        swingWorker.setState(state);
    }
    
    static void access$100(SwingWorker swingWorker) {
        swingWorker.doneEDT();
    }
    
    static {
        DO_SUBMIT_KEY = new StringBuilder("doSubmit");
    }
    
    public enum StateValue
    {
        PENDING, 
        STARTED, 
        DONE;
    }
    
    private static class DoSubmitAccumulativeRunnable extends AccumulativeRunnable<Runnable> implements ActionListener
    {
        private static final int DELAY = 33;
        
        @Override
        protected void run(List<Runnable> list) {
            Iterator<Object> iterator = list.iterator();
            while (iterator.hasNext()) {
                ((Runnable)iterator.next()).run();
            }
        }
        
        @Override
        protected void submit() {
            Timer timer = new Timer(33, this);
            timer.setRepeats(false);
            timer.start();
        }
        
        @Override
        public void actionPerformed(ActionEvent actionEvent) {
            super.run();
        }
    }
    
    private class SwingWorkerPropertyChangeSupport extends PropertyChangeSupport
    {
        SwingWorkerPropertyChangeSupport(Object sourceBean) {
            super(sourceBean);
        }
        
        @Override
        public void firePropertyChange(PropertyChangeEvent event) {
            if (SwingUtilities.isEventDispatchThread()) {
                super.firePropertyChange(event);
            }
            else {
                SwingWorker.this.doSubmit.add(new Runnable[] { new Runnable() {
                        @Override
                        public void run() {
                            SwingWorker.SwingWorkerPropertyChangeSupport.this.firePropertyChange(event);
                        }
                    } });
            }
        }
    }
}
